import json
import os
from os.path import exists

from typing import Dict

ALGO_FILE_NAME = "algo_spec.json"


class AlgoSpecHelper:
    __algo_spec = None
    __algo_spec_dir = None

    __recommender_name = None

    def __init__(self, recommender_name):
        self.__recommender_name = recommender_name
        self.__algo_spec_dir = os.path.dirname(os.path.realpath(__file__)) + "/repo/" + self.__recommender_name + "/"
        if exists(self.__algo_spec_dir + ALGO_FILE_NAME):
            self.load()

    def load(self):
        if not self.__algo_spec:
            with open(self.__algo_spec_dir + ALGO_FILE_NAME, "r") as openfile:
                self.__algo_spec = json.load(openfile)
        return self.__algo_spec

    def set(self, algo_spec=Dict):
        if algo_spec:
            os.makedirs(os.path.dirname(self.__algo_spec_dir), exist_ok=True)
            with open(self.__algo_spec_dir + ALGO_FILE_NAME, "w") as outfile:
                json.dump(algo_spec, outfile, indent=4)
                # json_data = json.dumps(algo_spec, indent=4)
                # outfile.write(json_data)
                self.__algo_spec = algo_spec
        else:
            raise FileNotFoundError("Algo spec can not be None")



import yodu
from examples.steem.helpers import load_test_items, load_test_actions
from models.request import Request

recommender = yodu.create_recommender(name="example")

recommender = yodu.get_recommender(name="example")

# Add Items to Recommender
items = load_test_items()
recommender.item.add(items)

# Add Actions to Recommender
actions = load_test_actions()
recommender.action.add(actions)

# Enable Yodu's built-in Providers
recommender.provider.add(name="top_item_by_user_action")
# Add provider from Source (Coming Soon)

algo_spec = {
    "TOP_BY_PREVIOUS_LIKED_SOURCES": {
        "provider": "top_item_by_user_action",
        "duration": "24h",
        "config": {
            "action_type": "LIKE",
            "tag": "source"
        },
        "weight": 1
    },
    "TOP_BY_PREVIOUS_LIKED_CATEGORIES": {
        "provider": "top_item_by_user_action",
        "duration": "30h",
        "config": {
            "action_type": "LIKE",
            "tag": "category"
        },
        "weight": 1
    },
    "TOP_BY_PREVIOUS_READ_CATEGORIES": {
        "provider": "top_item_by_user_action",
        "duration": "30h",
        "config": {
            "action_type": "READ",
            "tag": "category"
        },
        "weight": 1
    }
}
recommender.algo_spec.set(algo_spec=algo_spec)

args = {
    "days_ago": "7"
}
request = Request(user_id="test_user_1", args=args)

items = recommender.get_items(request=request)



import importlib
import os

from utils.utils import move_dir


class ProviderHelper:
    __recommender_name = None
    __provider_dir = None

    def __init__(self, recommender_name):
        self.__recommender_name = recommender_name
        self.__provider_dir = os.path.dirname(os.path.realpath(__file__))

    def add(self, name: str):
        '''
        Move from repo to recommender provider repo
        repo/recommender_name/enabled
        '''
        plugin_dir = "providers/" + name + "/"
        destination_dir = "repo/" + self.__recommender_name + "/" + name
        move_dir(self.__provider_dir, plugin_dir, destination_dir)

    def load_provider(self, name: str):
        print(self.__provider_dir)
        # import yodu.provider.repo.getting_started.top_item_by_user_action.provider
        provider_path = "yodu.provider.repo.getting_started." + name + ".provider"

        mod = importlib.import_module(provider_path)
        mod = getattr(mod, "Provider")
        return mod

    def remove(self, recommender_name: str, provider_name: str):
        '''
        Move from enabled to repo
        :return:
        '''
        pass

    def delete_provider(self):
        '''
        Check if provider_no_enabled
        Delete from repo
        :return:
        '''
        pass



from models.action import Action
from models.item import Item


def load_test_items():
    '''
    Create items with IDS: items0, item2...items99
    With categories ranging from category0...category9
    With source ranging from source0,...source9
    :return:
    '''
    items = []
    for i in range(0, 100):
        item = Item(
            id="item" + str(i),
            source="source" + str(i % 10),
            tags={
                "category": "category" + str(i % 10),
                "source": "source" + str(i % 10)
            })
        items.append(item)
    return items


def load_test_actions():
    actions = []
    for i in range(0, 100):
        action = Action(
            id="action" + str(i),
            user_id="user" + str(i % 10),
            item_id="item" + str(i),
            tags={
                "category": "category" + str(i % 10),
                "source": "source" + str(i % 10),
            },
            type="LIKE",
            value=1)
        actions.append(action)
    for i in range(0, 100):
        action = Action(
            id="action" + str(i),
            user_id="user" + str(i % 10),
            item_id="item" + str(i),
            tags={
                "category": "category" + str(i % 10),
                "source": "source" + str(i % 10),
            },
            type="READ",
            value=1)
        actions.append(action)
    return actions



import json
import os
import threading
from contextlib import suppress

from steem.blockchain import Blockchain


def get_last_line(filename):
    if os.path.isfile(filename):
        with open(filename, 'rb') as f:
            f.seek(-2, 2)
            while f.read(1) != b"\n":
                f.seek(-2, 1)
            return f.readline()


def get_previous_block_num(block):
    if not block:
        return -1

    if type(block) == bytes:
        block = block.decode('utf-8')

    if type(block) == str:
        block = json.loads(block)

    return int(block['previous'][:8], base=16)


def run(filename):
    b = Blockchain()
    # automatically resume from where we left off
    # previous + last + 1
    start_block = get_previous_block_num(get_last_line(filename)) + 2
    if start_block == 1:
        start_block = 66384718
    print("Block Number: " + str(start_block))
    with open(filename, 'a+') as file:
        for block in b.stream_from(start_block=start_block, full_blocks=False, batch_operations=True):
            file.write(json.dumps(block, sort_keys=True) + '\n')


def run_threaded(filename, start_block):
    b = Blockchain()
    with open(filename, 'a+') as file:
        for block in b.stream_from(start_block=start_block, full_blocks=False, batch_operations=True):
            file.write(json.dumps(block, sort_keys=True) + '\n')


if __name__ == '__main__':
    output_file = '/Users/shashank/PycharmProjects/yodu/data/steem.blockchain.json'
    with suppress(KeyboardInterrupt):
        run(output_file)



from elasticsearch import Elasticsearch

from yodu import settings


class ESClient:
    client = None

    def __int__(self):
        pass

    def get_client(self):
        if not self.client:
            es_uri = "https://" + settings.ES_USER + ":" + settings.ES_PASSWORD + "@" + settings.ES_HOST + ":" + str(
                settings.ES_PORT)
            self.client = Elasticsearch(hosts=es_uri, verify_certs=False)
        return self.client

    def get_docs(self, item_index, item_ids: list):
        hits = self.get_client().mget(index=item_index,
                                      body={
                                          'ids': item_ids
                                      })
        return hits



from influxdb_client import InfluxDBClient

from yodu import settings


class InfluxDb:
    client = None
    bucket = "bridgeml"

    def __init__(self, bucket="bridgeml"):
        self.bucket = bucket

    def get_client(self):
        if not self.client:
            self.client = InfluxDBClient(
                url=settings.INFLUX_DB_HOST,
                port=8086,
                username=settings.INFLUX_DB_USER,
                password=settings.INFLUX_DB_PASSWORD,
                ssl=settings.INFLUX_USE_SSL,
                verify_ssl=settings.INFLUX_USE_VERIFY_SSL,
            )
        return self.client

    def get_query_api(self):
        return self.get_client().query_api()

    def top_items_by_source(self, days_ago, action_type, tag_type, source, limit=5):
        """
        Return top items for a source by action_type
        Example: Return top liked items for a given source
        Example: Return top read items for a given source
        Example: Return top read items for a given category
        :param days_ago:
        :param action_type:
        :param source:
        :param limit:
        :return:
        """
        query = (
            'from(bucket: "bridgeml")\
        |> range(start: -'
            + str(days_ago)
            + 'd)\
        |> filter(fn: (r) => r["_measurement"] == "'
            + action_type
            + '")\
        |> filter(fn: (r) => r["'
            + tag_type
            + '"] == "'
            + source
            + '")\
        |> group(columns: ["item_id"])\
        |> aggregateWindow(every: '
            + str(days_ago)
            + 'd, fn: sum, createEmpty: false)\
        |> group()\
        |> sort(columns: ["_value"], desc: true)\
        |> top(n: '
            + str(limit)
            + ")"
        )
        res = self.get_query_api().query(org="influxdata", query=query)
        records = res[0].records
        top = {}
        for record in records:
            values = record.values
            top[values["item_id"]] = values["_value"]
        return top

    def top_by_user_action(self, days_ago, action_type, tag, user_id, limit=5):
        """
        Returns top tag for a user by action_type
        Ex: Top sources liked by the user
        Ex: Top categories liked by the user
        :param days_ago:
        :param action_type:
        :param tag:
        :param user_id:
        :param limit:
        :return:
        """
        query = (
            'from(bucket: "bridgeml")\
            |> range(start: -'
            + str(days_ago)
            + 'd)\
            |> filter(fn: (r) => r["_measurement"] == "'
            + action_type
            + '")\
            |> filter(fn: (r) => r["user_id"] == "'
            + str(user_id)
            + '")\
            |> group(columns: ["'
            + tag
            + '"])\
            |> aggregateWindow(every: '
            + str(days_ago)
            + 'd, fn: sum, createEmpty: false)\
            |> group()\
            |> sort(columns: ["_value"], desc: true) \
            |> top(n: '
            + str(limit)
            + ")"
        )
        res = self.get_query_api().query(org="influxdata", query=query)
        records = res[0].records
        top = {}
        for record in records:
            values = record.values
            top[values[tag]] = values["_value"]
        return top

    def top_by_action(self, days_ago, action_type, tag, limit=5):
        """
        Returns top tag for all items by action_type
        Ex: Top sources by likes
        :param days_ago:
        :param action_type:
        :param tag:
        :param limit:
        :return:
        """
        query = (
            'from(bucket: "bridgeml")\
            |> range(start: -'
            + str(days_ago)
            + 'd)\
            |> filter(fn: (r) => r["_measurement"] == "'
            + action_type
            + '")\
            |> group(columns: ["'
            + tag
            + '"])\
            |> aggregateWindow(every: '
            + str(days_ago)
            + 'd, fn: sum, createEmpty: false)\
            |> group()\
            |> sort(columns: ["_value"], desc: true) \
            |> top(n: '
            + str(limit)
            + ")"
        )
        res = self.get_query_api().query(org="influxdata", query=query)
        records = res[0].records
        top = {}
        for record in records:
            values = record.values
            top[values[tag]] = values["_value"]
        return top



from datetime import datetime
from typing import Optional, Dict
from uuid import uuid4

from pydantic import BaseModel, Field


class Item(BaseModel):
    id: str = Field(default_factory=uuid4)
    source: str
    tags: Optional[Dict]
    source_provider: str
    created_at: datetime = Field(default_factory=datetime.now)
    updated_at: datetime = Field(default_factory=datetime.now)
